
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */

/*
 *  $Id: ult.c,v 1.1 2002/09/02 00:56:47 steventoth Exp $
 *
 *  Copyright (c) 2002 Steven Toth <steve@toth.demon.co.uk>
 *
 *  This tool is a user space usb-xboxir.o configuration tool.
 *
 *  History:
 *
 *  2002_09_02 - 0.1 - Initial release
 *
 */

/*
 * Compile with:
 * gcc ult.c -o ult -I/usr/src/linux-2.4.19-XBOX-toths1/include -I/usr/src/linux-2.4.19-XBOX-toths1/drivers/usb
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/input.h>
#include <usb-xboxir.h>

#define PROGNAME "ult"
#define PROCNAME "/proc/xboxir"
#define VERSION  "0.1"

int fh,verbose=0;
struct t_rc_kbd_matrix entry;
char *src_name;
char *tgt_name;

char * get_rc_name_by_code(unsigned char c);
char * get_hid_name_by_code(unsigned int c);
void write_entry();
void usage(void);
void dump_proc(void);
unsigned int get_hid_key_by_name(char *n);
unsigned char get_rc_key_by_name(char *n);
void dump_keys(void);
char * get_rc_name_by_code(unsigned char c);
char * get_hid_name_by_code(unsigned int c);

void write_entry() {
	if (verbose) printf("Configuring driver: XBOX key %s generates kernel key %s\n", src_name, tgt_name);	
	if( (ioctl(fh,XBOXIR_IOCSQSET, &entry)) == -1) {
		printf("Error, unable to configure driver table\n");
		exit(1);
	}
}

void usage(void) {
	printf("Description: %s, a user space tool for configuring the xbox ir driver key codes.\n",PROGNAME);
	printf("Usage: %s [-hkq] -d /dev/name -s <src key> -t <target key>\n",PROGNAME);
	printf(" -h  help\n");
	printf(" -v  verbose\n");
	printf(" -V  version\n");
	printf(" -k  list key codes\n");
	printf(" -q  query current settings\n");
	printf(" -d  device name\n");
	printf(" -s  source keyname (XBOX)\n");
	printf(" -t  target keyname (KERNEL)\n\n");
	printf("Example: To remap the xbox INFO key to the HID stopcd KEY\n");
	printf("     do: %s -d /dev/xboxir -s RC_KEY_INFO -t KEY_STOPCD\n\n",PROGNAME);
}

void dump_proc(void) {
	struct t_rc_kbd_matrix entry;
	FILE *in;
	char line[80];
	int ret=0;
	

	in=fopen(PROCNAME,"rb");
	if (!in) {
		perror("fopen");
		exit(1);
	}

	printf("#XBOX key -> HID key\n");
	while(!feof(in)) {
		memset(&line,0,sizeof(line));
		fgets(&line[0],sizeof(line)-1,in);
		if(line[0] == '0') {
			ret=sscanf(line,"0x%02x:0x%04x",&entry.rc_code,&entry.kbd_code);
			//printf("ret=%d\n",ret);
			printf("%s -> %s\n",get_rc_name_by_code(entry.rc_code), get_hid_name_by_code(entry.kbd_code) );
		}
	}

	fclose(in);
}

/* For a given hid KEY_NAME, find it in the structures and return the unique code */
unsigned int get_hid_key_by_name(char *n) {
	int i=1; /* Start at 1 as 0 = reserved */

	while( hidkeys[i].code != 0 ) {
		if( strcmp(hidkeys[i].name,n) == 0) {
			return(hidkeys[i].code);
		}
		i++;
	}

	return 0;
}

void dump_keys(void) {
	int i=0;
	int w=0;

	printf("XBOX KEY NAME (CODE):\n");
	while( rckeys[i].code != 0 ) {
		printf("   %s (0x%x)\n",rckeys[i].name,rckeys[i].code);
		i++;
	}

	i=1; /* start from 1 as 0 is RESERVED */
	printf("HID KEY NAME (CODE):\n");
	while( hidkeys[i].code != 0 ) {
		printf("   %s (0x%x)\n",hidkeys[i].name,hidkeys[i].code);
		i++;
	}
}

/* For a given xbox KEY_NAME, find it in the structures and return the unique code */
unsigned char get_rc_key_by_name(char *n) {
	int i=0;

	while( rckeys[i].code != 0 ) {
		if( strcmp(rckeys[i].name,n) == 0) {
			return(rckeys[i].code);
		}
		i++;
	}

	return 0;
}

/* For a given xbox KEY_CODE, find it in the structures and return the name */
char * get_rc_name_by_code(unsigned char c) {
	int i=0;

	while( rckeys[i].code != 0 ) {
		if(rckeys[i].code == c) return(rckeys[i].name);
		i++;
	}

	return 0;
}

/* For a given hid KEY_CODE, find it in the structures and return the name */
char * get_hid_name_by_code(unsigned int c) {
	int i=1; /* Start at 1, 0 = reserved */

	while( hidkeys[i].code != 0 ) {
		if(hidkeys[i].code == c) return(hidkeys[i].name);
		i++;
	}

	return 0;
}


int main(int argc, char **argv[]) {

	extern char *optarg;
	extern int optind, opterr, optopt;
	int dflg=0,tflg=0,sflg=0,c=0;
	char dev[128],srckey[32],tgtkey[32];

	while ((c=getopt(argc,argv,"d:s:t:qvkhV")) != -1 ) {
		switch(c) {
			case 'V':
				printf("Version: %s\n", VERSION);
				exit(0);
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'k':
				dump_keys();
				exit(0);
				break;
			case 'q':
				dump_proc();
				exit(0);
				break;
			case 'v':
				verbose++;
				break;
			case 'd':
				dflg++;
				strcpy(dev,optarg);
				break;
			case 's':
				sflg++;
				if( (entry.rc_code = get_rc_key_by_name(optarg)) == 0) {
					printf("Error, invalid XBOX KEY_NAME\n");
					exit(1);
				}	
				src_name=optarg;
				break;
			case 't':
				tflg++;
				//entry.kbd_code = atoi(optarg);
				if( (entry.kbd_code = get_hid_key_by_name(optarg)) == 0) {
					printf("Error, invalid HID KEY_NAME\n");
					exit(1);
				}
				tgt_name=optarg;
				break;
			default:
				usage();
				exit(1);
		}
	}

	if(! ((dflg) && (sflg) && (tflg)) ) {
		usage();
		exit(1);
	}

	if (verbose) printf("Opening device driver named %s\n",dev);

	if( (fh=open(dev,O_RDWR)) == -1) {
		perror("open");
		exit(1);
	}

	write_entry();

	close(fh);
	if (verbose) printf("Closed device driver\n");
}

/*
#!/bin/sh
cat /usr/src/linux/include/linux/input.h | grep "#define" | grep "KEY_" | \
awk '{ print " {\"" $2 "\", " $2 "}," }'
*/

