/*
 *  fs/partitions/xbox.c
 *
 *  //TODO: find and insert GPL notice
 *
 *  June 2002 by SpeedBump:	initial implementation
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>

#ifdef CONFIG_BLK_DEV_IDE
#include <linux/ide.h>	/* IDE xlate */
#endif /* CONFIG_BLK_DEV_IDE */

#include <asm/system.h>

#include "check.h"
#include "xbox.h"

#define XBOX_SECTOR_STORE	(0x0055F400L)
#define XBOX_SECTOR_SYSTEM	(0x00465400L)
#define XBOX_SECTOR_CONFIG	(0x00000000L)
#define XBOX_SECTOR_CACHE1	(0x00000400L)
#define XBOX_SECTOR_CACHE2	(0x00177400L)
#define XBOX_SECTOR_CACHE3	(0x002EE400L)
//#define XBOX_SECTOR_EXTEND	(0x00EEEC00L)
#define XBOX_SECTOR_EXTEND	(0x00EE8AB0L)
#define XBOX_SECTORS_CONFIG	(XBOX_SECTOR_CACHE1 - XBOX_SECTOR_CONFIG)

#define XBOX_SECTORS_STORE	(XBOX_SECTOR_EXTEND - XBOX_SECTOR_STORE)
#define XBOX_SECTORS_SYSTEM	(XBOX_SECTOR_STORE  - XBOX_SECTOR_SYSTEM)
#define XBOX_SECTORS_CACHE1	(XBOX_SECTOR_CACHE2 - XBOX_SECTOR_CACHE1)
#define XBOX_SECTORS_CACHE2	(XBOX_SECTOR_CACHE3 - XBOX_SECTOR_CACHE2)
#define XBOX_SECTORS_CACHE3	(XBOX_SECTOR_SYSTEM - XBOX_SECTOR_CACHE3)
#define XBOX_SECTORS_CONFIG	(XBOX_SECTOR_CACHE1 - XBOX_SECTOR_CONFIG)

#define XBOX_SECTOR_MAGIC	(3L)

static int
xbox_sig_string_match(	struct block_device *bdev, 
			unsigned long at_sector,
			char *expect )
{
	Sector sect;
	int i = 0;
	int retv = 1;
	char *data;

	data = read_dev_sector(bdev, at_sector, &sect);
	
	if (!data) return 0;
	
	for(;(expect[i] != 0) && (i < 512); i++) {
		if (expect[i] != data[i]) {
			retv = 0;
			break;
		}
	}
	
	put_dev_sector(sect);
	
	/*
	if (!retv) {
		printk("xbox_sig_string_match: %s not found...found %c%c%c%c\n",
			expect,data[0],data[1],data[2],data[3]);
		for(i = 1; i<=512; i++) {
			printk(((i%32)?"%02X ":"%02X\n"),(unsigned char)data[i]);
		}
	}
	*/
	return retv;
}

static int
xbox_drive_protected( ide_drive_t *drive )
{
	ide_driveid_update(drive);
	
	return (drive->id->dlf & 2);
}
	
static int
xbox_drive_is_locked( ide_drive_t *drive )
{
	ide_driveid_update(drive);
	
	return (drive->id->dlf & (4+8+16));
}


static unsigned short ide_disk_security_password[16] = {
		0x5291,
		0x55C9,
		0x7C41,
		0xC97A,		//password, 16 words
		0x16C8,
		0xD7BB,
		0xDB2F,
		0xC126,
		0x6C2A,
		0x4920,
		0x0000,
		0x0000,
		0x0000,
		0x0000,
		0x0000,
		0x0000,		//rest is all reserved
	};

static void
xbox_unlock_drive( ide_drive_t *drive, int swap, unsigned short *password)
{
	//TODO: send the SECURITY UNLOCK command with password to drive
	struct request rq;
	unsigned char buffer[4+512] = { WIN_SECURITY_UNLOCK, 0, 0, 0, 0, 0 };
	
	memcpy(&(buffer[6]),password,32);
	
	if (swap) {
		int i;
		
		printk("byte-swapping password data\n");
		for(i = 0;i<32;i+=2) {
			unsigned char tmp = buffer[i+6];
			buffer[i+6] = buffer[i+7];
			buffer[i+7] = tmp;
		}
	}
	
	printk("xbox_unlock_drive: entered\n");
	mdelay(1000);
	
	ide_init_drive_cmd(&rq);
	
	rq.buffer = buffer;

	printk("xbox_unlock_drive: doing drive unlock command\n");
	mdelay(1000);	
	ide_do_drive_cmd(drive,&rq,ide_wait);
	
	printk("xbox_unlock_drive: completed\n");
	mdelay(1000);
}

static inline int
xbox_drive_detect(struct block_device *bdev)
{
	ide_drive_t *drive = get_info_ptr(to_kdev_t(bdev->bd_dev));
	
	/// detect if drive has security enabled and if it doesn't, fail.
	/* 27.09.2002: Changed by ED. It's also possible to have an
	 * unlocked HDD installd. If we only checking for locked
	 * HDD's, we can't see FATX partitions on unlocked DISKS
	 */
	if (!xbox_drive_protected(drive)) return 1;

	//for the moment, anything beyond here will report success
	//this prevents the other paritioning schemes from trying to read
	//from this (potentially) locked disk, generating a billion errors
		
	/// If it is protected, then unlock it and continue.
	if (xbox_drive_is_locked(drive)) {
		do {
			printk("drive is locked\n");
			xbox_unlock_drive(drive,0,ide_disk_security_password);
			//check unlocked status again, failing if still locked
			if (!xbox_drive_is_locked(drive)) {
				break;
			}
			printk("unlock failed -- trying swapped password\n");
			xbox_unlock_drive(drive,1,ide_disk_security_password);
			if (!xbox_drive_is_locked(drive)) {
				printk("unlock failed!\n");
				mdelay(1000);
				return -1;
			}
		} while (0);
		printk("unlocked!\n");
		mdelay(1000);
	}
	
	//we should now be able to read sectors from the drive...continue.
	
	/** 
	* "BRFR" is apparently the magic number in the config area
	* the others are just paranoid checks to assure the expected
	* "FATX" tags for the other xbox partitions
	*
	* the odds against a non-xbox drive having random data to match is
	* astronomical...but it's possible I guess...you should only include
	* this check if you actually *have* an xbox drive...since it has to
	* be detected first
	*
	* @see check.c
	*/
	if (	(xbox_sig_string_match(bdev,XBOX_SECTOR_MAGIC ,"BRFR")) &&
		(xbox_sig_string_match(bdev,XBOX_SECTOR_SYSTEM,"FATX")) &&
		(xbox_sig_string_match(bdev,XBOX_SECTOR_STORE ,"FATX"))) {
		return 1; //success
	}
	
	return -1; //failed
}

int xbox_partition(struct gendisk *hd, struct block_device *bdev,
		   unsigned long first_sector, int first_part_minor)
{
	ide_drive_t *drive = get_info_ptr(to_kdev_t(bdev->bd_dev));
	unsigned long last_sector = drive->capacity - 1;
	int minor = first_part_minor;
	int i = 0;
	int retv;
	
	if (first_sector != 0) {
		//we only accept whole ide drives...no partials
		printk("xbox_partition: failed...first_sector != 0 == %ld\n",first_sector);
		return 0;
	}
	
	retv = xbox_drive_detect(bdev);
	if (retv > 0) {
		/* trying to find the first free partition */
		minor = 50;
		add_gd_partition(hd,minor++,XBOX_SECTOR_STORE ,XBOX_SECTORS_STORE );
		add_gd_partition(hd,minor++,XBOX_SECTOR_SYSTEM,XBOX_SECTORS_SYSTEM);
		add_gd_partition(hd,minor++,XBOX_SECTOR_CACHE1,XBOX_SECTORS_CACHE1);
		add_gd_partition(hd,minor++,XBOX_SECTOR_CACHE2,XBOX_SECTORS_CACHE2);
		add_gd_partition(hd,minor++,XBOX_SECTOR_CACHE3,XBOX_SECTORS_CACHE3);
		//there are some questions about the usage of any "extra" sectors...
		//I'll map everything after the end to an aditional sector.
		if (XBOX_SECTOR_EXTEND < last_sector) {
			if (xbox_sig_string_match(bdev,XBOX_SECTOR_EXTEND ,"FATX")) {
				add_gd_partition(hd, 
					 minor++,
					 XBOX_SECTOR_EXTEND,
					 last_sector - XBOX_SECTOR_EXTEND);
			}
		}
	}
	
	if (retv == 0) {
		//not an xbox drive
		return 0;
	} else {
		printk("\n");
		//an xbox drive, though potentially locked (and, if so,
		//with no available partitions)
		return 1;
	}
}
